#include "../include/WordUtils.h"
#include <string>
#include <fstream>
#include <iostream>
#include <algorithm>
#include <cctype>
#include <cstring>
#include <cstdint>
#include <random>

map<string, map<string, int>> WordUtils::load_data(const string& path) {
    vector<string> data = split_string(load_file(path), data_type::HTML);

    map<string, map<string, int>> word_frequencies = {};

    for (long unsigned int i = 0; i + 1 < data.size(); i++) {
        transform(data[i].begin(), data[i].end(), data[i].begin(), [](unsigned char c){ return tolower(c); });
        transform(data[i+1].begin(), data[i+1].end(), data[i+1].begin(), [](unsigned char c){ return tolower(c); });
        word_frequencies[data[i]][data[i+1]]++;
    }

    return word_frequencies;
}

vector<string> WordUtils::load_css(const string& path) {
    return split_string(load_file(path), data_type::CSS);
}

vector<string> WordUtils::predict_next_word(const string& input, const map<string, map<string, int>>& word_frequencies, size_t count) {
    auto it = word_frequencies.find(input);

    if (it == word_frequencies.end()) return {input};

    const map<string, int> nextWords = it->second;

    vector<pair<string, int>> sortedWords(nextWords.begin(), nextWords.end());
    
    // Sort by frequency (descending)
    sort(sortedWords.begin(), sortedWords.end(),[](const auto& a, const auto& b) {
        return a.second > b.second;
    });

    vector<string> results;

    // Take up to "count" most common words
    for (size_t i = 0; i < min(count, sortedWords.size()); ++i) {
        results.push_back(sortedWords[i].first);
    }

    return results;
}

string WordUtils::create_tag(const map<string, map<string, int>>& word_frequencies, const char& hash) {
    unsigned char predict_num = 5;

    const string start_words[3] = {"the", "but", "with"};

    vector<string> tags;

    minstd_rand generator(hash);

    uniform_int_distribution<unsigned short> outer_distribution(0, start_words->length() - 1);
    
    vector<string> temp_words = predict_next_word(start_words[outer_distribution(generator)], word_frequencies, predict_num);

    uniform_int_distribution<unsigned short> outer_2_distribution(0, temp_words.size() - 1);

    tags.push_back(temp_words[outer_2_distribution(generator)]);

    // Words inside the <p> tag
    for (unsigned short j = 0; j < 25; j++)
    {
        temp_words = predict_next_word(tags[j], word_frequencies, predict_num);

        uniform_int_distribution<unsigned short> inner_distribution(0, temp_words.size() - 1);

        tags.push_back(temp_words[inner_distribution(generator)]);
    }

    string temp_string = "<p>";

    for (unsigned short l = 0; l < tags.size(); l++) {
        temp_string += tags[l];
        temp_string += " ";
    }

    temp_string += ".</p>\n";
    
    return temp_string;
}

vector<string> WordUtils::split_string(const string& input, data_type type) {
    vector<string> data;

    size_t start = 0;
    size_t end = 0;

    // Create a string from the delimiters array
    string delimiter_string;

    if (type == data_type::HTML)
    {
        delimiter_string = " .,!?;:()\n\r\t";
    }

    else if (type == data_type::CSS)
    {
        delimiter_string = "}\n\r\t";
    }
    
    
    while ((end = input.find_first_of(delimiter_string, start)) != string::npos) {
        if (end > start) {
            data.push_back(input.substr(start, end - start));
        }

        start = end + 1;
    }

    // Add the last token, if any
    if (start < input.length()) {
        data.push_back(input.substr(start));
    }

    return data;
}

string WordUtils::load_file(const string& path) {
    ifstream file(path);
    string temp;
    string data;

    while (getline(file, temp)) {
        data += temp;
    }

    file.close();

    return data;
}

string WordUtils::extract_url(const string& input) {
    unsigned short first_line_end = input.find("\n");
    if (first_line_end == string::npos) return "";

    string first_line = input.substr(0, first_line_end);

    unsigned short method_end = first_line.find(' ');
    if (method_end == string::npos) return "";

    unsigned short path_end = first_line.find(' ', method_end + 1);
    if (path_end == string::npos) return "";

    return first_line.substr(method_end + 1, path_end - method_end - 1);
}

unsigned int WordUtils::hash_url(const string& input) {
    unsigned int hash = 0;

    for (char c : input) {
        hash += static_cast<unsigned int>(c) * 597301;
    }

    return hash;
}